#  Copyright 2022 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# Standard
from typing import Callable, Optional

# Third Party
import numpy as np
import pandas as pd
import tensorflow as tf

# Rikai
from rikai.parquet.dataset import convert_tensor
from rikai.spark.sql.codegen.base import unpickle_transform

__all__ = ["PandasDataset"]


class PandasDataset:
    """a Map-style Tensorflow dataset from a :py:class:`pandas.DataFrame` or a
    :py:class:`pandas.Series`.

    Note
    ----

    This class is used in Rikai's SQL-ML Spark implementation, which utilizes
    pandas UDF to run inference.
    """

    def __init__(
        self,
        df: pd.Series,
        transform: Optional[Callable] = None,
        unpickle: bool = False,
        use_pil: bool = False,
    ) -> None:
        assert isinstance(df, pd.Series)
        self.df = df
        self.transform = transform
        self.unpickle = unpickle
        self.use_pil = use_pil

    def batch(self, batch_size):
        def upickle_convent_transform(entity):
            if self.unpickle:
                entity = unpickle_transform(entity)
            entity = convert_tensor(entity, use_pil=self.use_pil)

            from rikai.types.vision import Image

            ret = Image.from_pil(entity).to_numpy()
            if self.transform:
                ret = self.transform(ret)
            return ret

        # TODO I want to use
        # `tensors = self.df.map(upickle_convent_transform).to_numpy()`
        # here, but it will cause unintelligible error
        tensors = np.array([upickle_convent_transform(x) for x in self.df])

        data = tf.data.Dataset.from_tensor_slices(tensors)
        # data = tf.data.Dataset.from_tensors(img)

        data = data.batch(batch_size).as_numpy_iterator()
        return data



#  Copyright 2022 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

from typing import Any, Callable, Optional

import torch
import torchvision

from rikai.pytorch.models.torch import TorchModelType
from rikai.spark.sql.model import ModelSpec, parse_model_type

__all__ = ["FeatureExtractor", "FeatureExtractorType"]


class FeatureExtractor(torch.nn.Module):
    """Extract features"""

    def __init__(
        self,
        model: torch.nn.Module,
        node: str,
        output_field: str = "_rikai_out",
    ):
        super().__init__()
        self._node = node

        from torchvision.models.feature_extraction import (
            create_feature_extractor,
        )

        self.model = create_feature_extractor(model, {node: output_field})
        self._output_field = output_field

    def eval(self: FeatureExtractor) -> FeatureExtractor:
        self.model.eval()
        return self

    def forward(self, images: torch.Tensor) -> torch.Tensor:
        preds = self.model(images)
        return preds[self._output_field]


class FeatureExtractorType(TorchModelType):
    """(Experimental) Generic Feature Extractor

    Examples
    --------

    .. code-block:: sql

        CREATE MODEL resnet_features
        FLAVOR pytorch
        MODEL_TYPE feature_extractors
        OPTIONS (model_type = 'resnet')  # Only resnet is tested
        USING '<uri to resnet model>'

        SELECT ML_PREDICT(resnet_features, image) AS embedding FROM images
    """

    def __init__(self):
        super().__init__("feature_extractor")
        self.original_model_type: Optional[TorchModelType] = None

    def load_model(self, spec: ModelSpec, **kwargs):
        self.spec = spec
        self.model = self.spec.load_model()
        self.original_model_type = parse_model_type(
            "pytorch", self.spec.options["model_type"]
        )
        self.model.eval()
        if "device" in kwargs:
            self.model.to(kwargs.get("device"))

    def schema(self) -> str:
        return "array<float>"

    def transform(self) -> Callable:
        assert (
            self.original_model_type is not None
        ), "The original model type has not been initialized"
        return self.original_model_type.transform()

    def predict(self, *args, **kwargs) -> Any:
        out = self.model(*args, **kwargs)
        batch = []
        for row in out.cpu():
            batch.append(row.T[0, 0].tolist())
        return batch


if torchvision.__version__ >= "0.11.0":
    feature_extractor = FeatureExtractorType()



#  Copyright 2022 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

#  Copyright 2022 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

from abc import ABC
from typing import Callable, List

from rikai.spark.sql.model import ModelSpec, ModelType


class SklearnModelType(ModelType, ABC):
    """Base :py:class:`ModelType` for Sklearn"""

    def __init__(self):
        self.model = None

    def load_model(self, spec: ModelSpec, **kwargs):
        self.model = spec.load_model()

    def transform(self) -> Callable:
        # Do nothing in Sklearn
        pass


class Classification(SklearnModelType):
    """Classification model type"""

    def schema(self) -> str:
        return "int"

    def predict(self, x, *args, **kwargs) -> dict:
        assert self.model is not None
        return self.model.predict(x).tolist()


class Regression(SklearnModelType):
    def schema(self) -> str:
        return "float"

    def predict(self, x, *args, **kwargs) -> float:
        assert self.model is not None
        return self.model.predict(x).tolist()


class DimensionalityReduction(SklearnModelType):
    """Dimensionality reduction models.

    - PCA
    """

    def schema(self) -> str:
        return "array<float>"

    def predict(self, x, *args, **kwargs) -> List[float]:
        return self.model.transform(x).tolist()


MODEL_TYPES = {
    "linear_regression": Regression(),
    "random_forest_regression": Regression(),
    "logistic_regression": Classification(),
    "random_forest_classification": Classification(),
    "pca": DimensionalityReduction(),
}



#  Copyright 2022 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
EfficientNet: Rethinking Model Scaling for Convolutional Neural Networks
https://arxiv.org/abs/1905.11946
"""

import torchvision

from rikai.pytorch.models.torch import (
    ClassificationModelType,
)

if torchvision.__version__ >= "0.11.0":
    efficientnet_b0 = ClassificationModelType(
        "efficientnet_b0", pretrained_fn=torchvision.models.efficientnet_b0
    )
    efficientnet_b1 = ClassificationModelType(
        "efficientnet_b1", pretrained_fn=torchvision.models.efficientnet_b1
    )
    efficientnet_b2 = ClassificationModelType(
        "efficientnet_b2", pretrained_fn=torchvision.models.efficientnet_b2
    )
    efficientnet_b3 = ClassificationModelType(
        "efficientnet_b3", pretrained_fn=torchvision.models.efficientnet_b3
    )
    efficientnet_b4 = ClassificationModelType(
        "efficientnet_b4", pretrained_fn=torchvision.models.efficientnet_b4
    )
    efficientnet_b5 = ClassificationModelType(
        "efficientnet_b5", pretrained_fn=torchvision.models.efficientnet_b5
    )
    efficientnet_b6 = ClassificationModelType(
        "efficientnet_b6", pretrained_fn=torchvision.models.efficientnet_b6
    )
    efficientnet_b7 = ClassificationModelType(
        "efficientnet_b7", pretrained_fn=torchvision.models.efficientnet_b7
    )



#  Copyright 2022 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import torchvision

from rikai.pytorch.models.torch import ObjectDetectionModelType


def label_fn(label_id):
    # https://pytorch.org/vision/stable/models.html
    COCO_PERSON_KEYPOINT_NAMES = [
        "nose",
        "left_eye",
        "right_eye",
        "left_ear",
        "right_ear",
        "left_shoulder",
        "right_shoulder",
        "left_elbow",
        "right_elbow",
        "left_wrist",
        "right_wrist",
        "left_hip",
        "right_hip",
        "left_knee",
        "right_knee",
        "left_ankle",
        "right_ankle",
    ]
    return COCO_PERSON_KEYPOINT_NAMES[label_id]


keypointrcnn = ObjectDetectionModelType(
    "keypointrcnn",
    pretrained_fn=torchvision.models.detection.keypointrcnn_resnet50_fpn,
    label_fn=label_fn,
)



#  Copyright 2022 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Rikai-implemented PyTorch models and executors."""

import importlib

torchvision_found = importlib.util.find_spec("torchvision") is not None

if torchvision_found:
    import rikai.pytorch.models.convnext
    import rikai.pytorch.models.efficientnet
    import rikai.pytorch.models.fasterrcnn
    import rikai.pytorch.models.feature_extractor
    import rikai.pytorch.models.keypointrcnn
    import rikai.pytorch.models.maskrcnn
    import rikai.pytorch.models.resnet
    import rikai.pytorch.models.retinanet
    import rikai.pytorch.models.ssd
    import rikai.pytorch.models.ssd_class_scores

from rikai.pytorch.models.torch import MODEL_TYPES  # noqa

__all__ = ["MODEL_TYPES"]
